from __future__ import annotations

from dataclasses import dataclass
from enum import auto
from typing import List, Literal, Union
from uuid import UUID

from encord.http.v2.api_client import ApiClient
from encord.orm.analytics import CamelStrEnum
from encord.orm.base_dto import BaseDTO


class _IssueAnchorType(CamelStrEnum):
    DATA_UNIT = auto()
    FRAME = auto()
    FRAME_COORDINATE = auto()


class _FileIssueAnchor(BaseDTO):
    type: Literal[_IssueAnchorType.DATA_UNIT] = _IssueAnchorType.DATA_UNIT
    data_uuid: UUID


class _FrameIssueAnchor(BaseDTO):
    type: Literal[_IssueAnchorType.FRAME] = _IssueAnchorType.FRAME
    data_uuid: UUID
    frame_index: int


class _CoordinateIssueAnchor(BaseDTO):
    type: Literal[_IssueAnchorType.FRAME_COORDINATE] = _IssueAnchorType.FRAME_COORDINATE
    data_uuid: UUID
    frame_index: int
    x: float
    y: float


_IssueAnchor = Union[_FileIssueAnchor, _FrameIssueAnchor, _CoordinateIssueAnchor]


class _NewIssue(BaseDTO):
    anchor: _IssueAnchor
    comment: str
    issue_tags: List[str]


class _CreateIssuesPayload(BaseDTO):
    issues: List[_NewIssue]


class _IssueClient:
    def __init__(self, api_client: ApiClient) -> None:
        self._api_client = api_client

    def add_issue(self, project_uuid: UUID, anchor: _IssueAnchor, comment: str, issue_tags: List[str]) -> None:
        self._api_client.post(
            path=f"/projects/{project_uuid}/issues",
            params=None,
            payload=_CreateIssuesPayload(
                issues=[
                    _NewIssue(
                        anchor=anchor,
                        comment=comment,
                        issue_tags=issue_tags,
                    )
                ]
            ),
            result_type=None,
        )


class TaskIssues:
    def __init__(self, api_client: ApiClient, project_uuid: UUID, data_uuid: UUID):
        self._issue_client = _IssueClient(api_client=api_client)
        self._project_uuid = project_uuid
        self._data_uuid = data_uuid

    def add_file_issue(self, comment: str, issue_tags: List[str]) -> None:
        """Adds a file issue.

        Args:
            comment (str): The comment for the issue.
            issue_tags (List[str]): The issue tags for the issue.
        """
        self._issue_client.add_issue(
            project_uuid=self._project_uuid,
            anchor=_FileIssueAnchor(
                data_uuid=self._data_uuid,
            ),
            comment=comment,
            issue_tags=issue_tags,
        )

    def add_frame_issue(self, frame_index: int, comment: str, issue_tags: List[str]) -> None:
        """Adds a frame issue.

        Args:
            frame_index (int): The index of the frame to add the issue to.
            comment (str): The comment for the issue.
            issue_tags (List[str]): The issue tags for the issue.
        """
        self._issue_client.add_issue(
            project_uuid=self._project_uuid,
            anchor=_FrameIssueAnchor(data_uuid=self._data_uuid, frame_index=frame_index),
            comment=comment,
            issue_tags=issue_tags,
        )

    def add_coordinate_issue(self, frame_index: int, x: float, y: float, comment: str, issue_tags: List[str]) -> None:
        """Adds a issue pinned to a coordinate.

        Args:
            frame_index (int): The index of the frame to add the issue to.
            x (float): The x coordinate of the issue.
            y (float): The y coordinate of the issue.
            comment (str): The comment for the issue.
            issue_tags (List[str]): The issue tags for the issue.
        """
        self._issue_client.add_issue(
            project_uuid=self._project_uuid,
            anchor=_CoordinateIssueAnchor(
                data_uuid=self._data_uuid,
                frame_index=frame_index,
                x=x,
                y=y,
            ),
            comment=comment,
            issue_tags=issue_tags,
        )



from encord.objects.attributes import ChecklistAttribute, RadioAttribute, TextAttribute
from encord.objects.classification import Classification
from encord.objects.classification_instance import ClassificationInstance
from encord.objects.common import Shape
from encord.objects.metadata import DICOMSeriesMetadata, DICOMSliceMetadata
from encord.objects.ontology_labels_impl import LabelRowV2
from encord.objects.ontology_object import Object
from encord.objects.ontology_object_instance import AnswerForFrames, ObjectInstance
from encord.objects.ontology_structure import OntologyStructure
from encord.objects.options import FlatOption, NestableOption, Option

__all__ = [
    "AnswerForFrames",
    "ChecklistAttribute",
    "Classification",
    "ClassificationInstance",
    "DICOMSeriesMetadata",
    "DICOMSliceMetadata",
    "FlatOption",
    "LabelRowV2",
    "NestableOption",
    "Object",
    "ObjectInstance",
    "OntologyStructure",
    "Option",
    "RadioAttribute",
    "Shape",
    "TextAttribute",
]



from datetime import datetime
from typing import Iterator, List, Optional, Union
from uuid import UUID

from encord.client import EncordClientProject
from encord.exceptions import (
    AuthorisationError,
    EncordException,
)
from encord.http.v2.api_client import ApiClient
from encord.orm.filter_preset import (
    ActiveCreatePresetPayload,
    ActiveFilterPresetDefinition,
    ActiveUpdatePresetPayload,
    CreatePresetParams,
    GetPresetParams,
    GetPresetsResponse,
    GetProjectFilterPresetParams,
    IndexCreatePresetPayload,
    IndexFilterPresetDefinition,
    IndexUpdatePresetPayload,
)
from encord.orm.filter_preset import FilterPreset as OrmFilterPreset
from encord.orm.filter_preset import ProjectFilterPreset as OrmProjectFilterPreset


class FilterPreset:
    """Represents preset in Index.
    Preset is a group of filters persisted which can be reused for faster data curation.
    """

    def __init__(self, client: ApiClient, orm_preset: OrmFilterPreset):
        self._client = client
        self._preset_instance = orm_preset

    @property
    def uuid(self) -> UUID:
        """Get the preset uuid (i.e. the preset ID).

        Returns:
            str: The preset uuid.
        """
        return self._preset_instance.uuid

    @property
    def name(self) -> str:
        """Get the preset name

        Returns:
            str: The preset name.
        """
        return self._preset_instance.name

    @property
    def description(self) -> Optional[str]:
        """Get the preset description

        Returns:
            Optional[str]: The preset description.
        """
        return self._preset_instance.description

    @property
    def created_at(self) -> Optional[datetime]:
        """Get the preset creation timestamp

        Returns:
            Optional[datetime]: The preset creation timestamp.
        """
        return self._preset_instance.created_at

    @property
    def last_updated_at(self) -> Optional[datetime]:
        """Get the preset last update timestamp

        Returns:
            Optional[datetime]: The preset last update timestamp.
        """
        return self._preset_instance.last_updated_at

    @staticmethod
    def _get_preset(api_client: ApiClient, preset_uuid: UUID) -> "FilterPreset":
        params = GetPresetParams(uuids=[preset_uuid])
        orm_item = api_client.get(
            "index/presets",
            params=params,
            result_type=GetPresetsResponse,
        )
        if len(orm_item.results) > 0:
            return FilterPreset(api_client, orm_item.results[0])
        raise AuthorisationError("Preset not found")

    @staticmethod
    def _get_presets(
        api_client: ApiClient,
        preset_uuids: List[UUID],
        page_size: Optional[int] = None,
    ) -> Iterator["FilterPreset"]:
        params = GetPresetParams(uuids=preset_uuids, pageSize=page_size)
        paged_items = api_client.get_paged_iterator(
            "index/presets",
            params=params,
            result_type=OrmFilterPreset,
        )
        for item in paged_items:
            yield FilterPreset(api_client, item)

    @staticmethod
    def _list_presets(
        api_client: ApiClient, top_level_folder_uuid: Union[UUID, None], page_size: Optional[int] = None
    ) -> Iterator["FilterPreset"]:
        params = GetPresetParams(topLevelFolderUuid=top_level_folder_uuid, pageSize=page_size)
        paged_items = api_client.get_paged_iterator(
            "index/presets",
            params=params,
            result_type=OrmFilterPreset,
        )
        for item in paged_items:
            yield FilterPreset(api_client, item)

    @staticmethod
    def _delete_preset(api_client: ApiClient, preset_uuid: UUID) -> None:
        api_client.delete(
            f"index/presets/{preset_uuid}",
            params=None,
            result_type=None,
        )

    @staticmethod
    def _create_preset(api_client: ApiClient, name: str, description: str = "", *, filter_preset_json: dict) -> UUID:
        filter_preset = IndexFilterPresetDefinition.from_dict(filter_preset_json)
        if not filter_preset.local_filters and not filter_preset.global_filters:
            raise EncordException("We require there to be a non-zero number of filters in a preset")
        payload = IndexCreatePresetPayload(
            name=name,
            filter_preset_json=filter_preset.to_dict(),
            description=description,
        )
        return api_client.post(
            "index/presets",
            payload=payload,
            params=CreatePresetParams(),
            result_type=UUID,
            allow_retries=False,
        )

    def get_filter_preset_json(self) -> IndexFilterPresetDefinition:
        return self._client.get(
            f"index/presets/{self._preset_instance.uuid}",
            params=None,
            result_type=IndexFilterPresetDefinition,
        )

    def update_preset(
        self, name: Optional[str] = None, description: Optional[str] = None, filter_preset_json: Optional[dict] = None
    ) -> None:
        """Update the preset's definition.

        Args:
           name (Optional[str]): The new name for the preset.
           description (Optional[str]): The new description for the preset.
           filter_preset_json (Optional[dict]): The new filters for the preset in their raw json format.
        """
        filters_definition = None
        if isinstance(filter_preset_json, dict):
            filters_definition = IndexFilterPresetDefinition.from_dict(filter_preset_json)
        elif isinstance(filter_preset_json, IndexFilterPresetDefinition):
            filters_definition = filter_preset_json
        if filters_definition:
            if not filters_definition.local_filters and not filters_definition.global_filters:
                raise EncordException("We require there to be a non-zero number of filters in a preset")
        payload = IndexUpdatePresetPayload(name=name, description=description, filter_preset=filters_definition)
        self._client.patch(
            f"index/presets/{self.uuid}",
            params=None,
            payload=payload,
            result_type=None,
        )
        self._preset_instance.name = name or self.name
        self._preset_instance.description = description or self.description


class ProjectFilterPreset:
    """Represents Active filter presets."""

    def __init__(
        self,
        project_uuid: UUID,
        client: ApiClient,
        orm_filter_preset: OrmProjectFilterPreset,
    ):
        self._project_uuid = project_uuid
        self._client = client
        self._filter_preset_instance = orm_filter_preset

    @property
    def uuid(self) -> UUID:
        """Get the filter preset unique identifier (UUID).

        Returns:
            UUID: The filter preset UUID.
        """
        return self._filter_preset_instance.preset_uuid

    @property
    def name(self) -> str:
        """Get the filter preset name.

        Returns:
            str: The collection name.
        """
        return self._filter_preset_instance.name

    @property
    def created_at(self) -> Optional[datetime]:
        """Get the filter preset creation timestamp.

        Returns:
            Optional[datetime]: The timestamp when the filter preset was created, or None if not available.
        """
        return self._filter_preset_instance.created_at

    @property
    def updated_at(self) -> Optional[datetime]:
        """Get the filter preset last edit timestamp.

        Returns:
            Optional[datetime]: The timestamp when the filter preset was last edited, or None if not available.
        """
        return self._filter_preset_instance.updated_at

    @property
    def project_hash(self) -> UUID:
        """Get the project hash of the filter preset.

        Returns:
            UUID: The project hash of the filter preset.
        """
        return self._project_uuid

    @staticmethod
    def _get_filter_preset(
        client: ApiClient,
        project_uuid: UUID,
        filter_preset_uuid: UUID,
    ) -> "ProjectFilterPreset":
        params = GetProjectFilterPresetParams(preset_uuids=[filter_preset_uuid])
        orm_items = list(
            client.get_paged_iterator(
                f"active/{project_uuid}/presets",
                params=params,
                result_type=OrmProjectFilterPreset,
            )
        )
        if len(orm_items) > 0:
            return ProjectFilterPreset(
                project_uuid=project_uuid,
                client=client,
                orm_filter_preset=orm_items[0],
            )
        raise AuthorisationError("No Project preset found")

    @staticmethod
    def _list_filter_presets(
        client: ApiClient,
        project_uuid: UUID,
        filter_preset_uuids: Union[List[UUID], None],
        page_size: Optional[int] = None,
    ) -> Iterator["ProjectFilterPreset"]:
        params = GetProjectFilterPresetParams(preset_uuids=filter_preset_uuids, page_size=page_size)
        paged_filter_presets = client.get_paged_iterator(
            f"active/{project_uuid}/presets",
            params=params,
            result_type=OrmProjectFilterPreset,
        )
        for filter_preset in paged_filter_presets:
            yield ProjectFilterPreset(
                project_uuid=project_uuid,
                client=client,
                orm_filter_preset=filter_preset,
            )

    @staticmethod
    def _delete_filter_preset(client: ApiClient, project_uuid: UUID, filter_preset_uuid: UUID) -> None:
        client.delete(
            f"active/{project_uuid}/presets/{filter_preset_uuid}",
            params=None,
            result_type=None,
        )

    def get_filter_preset_json(self) -> ActiveFilterPresetDefinition:
        return self._client.get(
            f"active/{self._project_uuid}/presets/{self._filter_preset_instance.preset_uuid}/raw",
            params=None,
            result_type=ActiveFilterPresetDefinition,
        )

    def update_preset(
        self, name: Optional[str] = None, filter_preset: Optional[ActiveFilterPresetDefinition] = None
    ) -> None:
        if name is None and filter_preset is None:
            return
        payload = ActiveUpdatePresetPayload(name=name, filter_preset=filter_preset)
        self._client.patch(
            f"active/{self.project_hash}/presets/{self.uuid}", params=None, payload=payload, result_type=None
        )

    @staticmethod
    def _create_filter_preset(
        client: ApiClient, project_uuid: UUID, name: str, filter_preset: ActiveFilterPresetDefinition
    ) -> UUID:
        if not filter_preset.local_filters and not filter_preset.global_filters:
            raise EncordException("We require there to be a non-zero number of filters in a preset for creation")
        payload = ActiveCreatePresetPayload(name=name, filter_preset_json=filter_preset.to_dict())
        orm_resp = client.post(
            f"active/{project_uuid}/presets",
            params=None,
            payload=payload,
            result_type=UUID,
            allow_retries=False,
        )
        return orm_resp



from __future__ import annotations

import datetime
from collections import OrderedDict
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import Field

from encord.common.time_parser import parse_datetime
from encord.orm import base_orm
from encord.orm.base_dto import BaseDTO
from encord.orm.formatter import Formatter


class LabelRow(base_orm.BaseORM):
    """A label row contains a data unit or a collection of data units and associated
    labels, and is specific to a data asset with type ``video`` or ``img_group``:

    * A label row with a data asset of type video contains a single data unit.
    * A label row with a data asset of type img_group contains any number of data units.

    The label row ORM is as follows:

    * ``label_hash`` (uid) is the unique identifier of the label row
    * ``dataset_hash`` (uid) is the unique identifier of the dataset which contains the
      particular video or image group
    * ``dataset_title`` is the title of the dataset which contains the particular video
      or image group
    * ``data_title`` is the title of the video or image group
    * ``data_type`` either ``video`` or ``img_group`` depending on data type
    * ``data_units`` a dictionary with (key: data hash, value: data unit) pairs.
    * ``object_answers`` is a dictionary with (key: object hash, value: object answer)
      pairs.
    * ``classification_answers`` is a dictionary with (key: classification hash, value:
      classification answer) pairs.
    * ``object_actions`` is a dictionary with (key: ``<object_hash>``, value: object
      action) pairs.
    * ``label_status`` is a string indicating label status. It can take the values
      enumerated in :class:`encord.orm.label_row.LabelStatus`. *Note* that this does
      *not* reflect thes status shown in the Projects->Labels section on the web-app.

    A data unit, mentioned for the dictionary entry ``data_units`` above, has in the
    form::

        label_row = {  # The label row
            # ...
            "data_units": {
                "<data_hash>": {
                    "data_hash": "<data_hash>",  # A data_hash (uid) string
                    "data_title": "A data title",
                    "data_link": "<data_link>",  # Signed URL that expiring after 7 days
                    "data_type": "<data_type>",  # (video/mp4, image/jpeg, etc.)
                    "data_fps": 24.95,           # For video, the frame rate
                    "data_sequence": "0",        # Defines order of data units
                    "width": 640,                # The width of the content
                    "height": 610,               # The height of the content
                    "labels": {
                        # ...
                    }
                },
                # ...,
            }
        }

    A data unit can have any number of vector labels (e.g. bounding box, polygon, keypoint) and classifications.

    **Objects and classifications**

    A data unit can have any number of vector labels (e.g., bounding boxes, polygons,
    polylines, keypoints) and classifications.
    Each frame-level object and classification has unique identifiers 'objectHash' and
    'classificationHash'. Each frame-level entity has a unique feature identifier
    'featureHash', defined in the editor ontology.

    The object and classification answers are contained separately from the individual
    data units to preserve space for video, sequential images, DICOM, etc.

    The objects and classifications answer dictionaries contain classification 'answers'
    (i.e. attributes that describe the object or classification). This is to avoid
    storing the information at every frame in the blurb, of particular importance for
    videos.

    A labels dictionary for video is in the form::

        label_row["data_units"]["<data_hash>"]["labels"] = {
            "<frame_number>": {
                "objects": [
                    # { object 1 },
                    # { object 2 },
                    # ...
                ],
                "classifications": [
                    # { classification 1 },
                    # { classification 2 },
                    # ...
                ],
            }
        }

    A labels dictionary for an img_group data unit is in the form::

        label_row["data_units"]["<data_hash>"]["labels"] = {
            "objects": [
                # { object 1 },
                # { object 2 },
                # ...
            ],
            "classifications": [
                # { classification 1 },
                # { classification 2 },
                # ...
            ],
        }

    The object answers dictionary is in the form::

        label_row["object_answers"] = {
            "<object_hash>": {
                "objectHash": "<object_hash>",
                "classifications": [
                    # {answer 1},
                    # {answer 2},
                    # ...
                ]
            },
            # ...
        }

    The classification answers dictionary is in the form::

        label_row["classification_answers"] = {
            "<classification_hash>": {
                "classificationHash": "<classification_hash>",
                "classifications": [
                    # {answer 1},
                    # {answer 2},
                    # ...
                ],
            },
            # ...
        }

    The object actions dictionary is in the form::

        label_row["object_actions"] = {
            "<object_hash>": {
                "objectHash": "<object_hash>",
                "actions": [
                    # {answer 1},
                    # {answer 2},
                    # ...
                ],
            },
            # ...
        }

    """

    DB_FIELDS = OrderedDict(
        [
            ("label_hash", str),
            ("branch_name", str),
            ("created_at", str),
            ("last_edited_at", str),
            ("dataset_hash", str),
            ("dataset_title", str),
            ("data_title", str),
            ("data_hash", str),
            ("data_type", str),
            ("data_units", dict),
            ("object_answers", dict),
            ("classification_answers", dict),
            ("object_actions", dict),
            ("label_status", str),
            ("annotation_task_status", str),
            ("is_valid", bool),
        ]
    )

    NON_UPDATABLE_FIELDS = {
        "label_hash",
    }


class AnnotationTaskStatus(Enum):
    QUEUED = "QUEUED"
    ASSIGNED = "ASSIGNED"
    IN_REVIEW = "IN_REVIEW"
    RETURNED = "RETURNED"
    COMPLETED = "COMPLETED"


class ShadowDataState(Enum):
    """Specifies the kind of data to fetch when working with a BenchmarkQa project"""

    ALL_DATA = "ALL_DATA"
    """ Fetch all the label rows """
    SHADOW_DATA = "SHADOW_DATA"
    """ Only fetch the label rows that were submitted against "shadow data": the annotator's view of the benchmark """
    NOT_SHADOW_DATA = "NOT_SHADOW_DATA"
    """ Only fetch the label rows for "production" data """


class LabelStatus(Enum):
    NOT_LABELLED = "NOT_LABELLED"
    LABEL_IN_PROGRESS = "LABEL_IN_PROGRESS"
    LABELLED = "LABELLED"
    REVIEW_IN_PROGRESS = "REVIEW_IN_PROGRESS"
    REVIEWED = "REVIEWED"
    REVIEWED_TWICE = "REVIEWED_TWICE"

    MISSING_LABEL_STATUS = "_MISSING_LABEL_STATUS_"
    """
    This value will be displayed if the Encord platform has a new label status and your SDK version does not understand
    it yet. Please update your SDK to the latest version.
    """

    @classmethod
    def _missing_(cls, value: Any) -> LabelStatus:
        return cls.MISSING_LABEL_STATUS


@dataclass(frozen=True)
class WorkflowGraphNode:
    uuid: str
    title: str

    @classmethod
    def from_optional_dict(cls, json_dict: Optional[Dict]) -> Optional[WorkflowGraphNode]:
        if json_dict is None:
            return None
        return WorkflowGraphNode(uuid=json_dict["uuid"], title=json_dict["title"])


@dataclass(frozen=True)
class LabelRowMetadata(Formatter):
    """Contains helpful information about a label row."""

    label_hash: Optional[str]
    """Only present if the label row is initiated"""
    created_at: Optional[datetime.datetime]
    """Only present if the label row is initiated"""
    last_edited_at: Optional[datetime.datetime]
    """Only present if the label row is initiated"""
    branch_name: str
    """Only present if the label row is initiated or branch_name is set specifically"""

    data_hash: str
    """Only present if the label row is for a child element of a group"""
    dataset_hash: str
    dataset_title: str
    data_title: str
    data_type: str
    data_link: Optional[str]
    """Can be `None` for label rows of image groups or DICOM series."""
    label_status: LabelStatus
    """Can be `None` for TMS2 projects"""
    annotation_task_status: Optional[AnnotationTaskStatus]
    """Only available for TMS2 project"""
    workflow_graph_node: Optional[WorkflowGraphNode]
    is_shadow_data: bool

    """Only available for the VIDEO and AUDIO data_type"""
    frames_per_second: Optional[float]
    number_of_frames: int
    duration: Optional[float]

    """Only available for the VIDEO data_type"""
    height: Optional[int]
    width: Optional[int]

    """Only available for the AUDIO data_type"""
    audio_sample_rate: Optional[int]
    audio_codec: Optional[str]
    audio_bit_depth: Optional[int]
    audio_num_channels: Optional[int]

    group_hash: Optional[str] = None

    task_uuid: Optional[UUID] = None
    priority: Optional[float] = None
    """Only available for not complete tasks"""
    client_metadata: Optional[dict] = None
    images_data: Optional[list] = None
    file_type: Optional[str] = None
    """Only available for certain read requests"""
    is_valid: bool = True

    backing_item_uuid: Optional[UUID] = None
    assigned_user_email: Optional[str] = None
    last_actioned_by_user_email: Optional[str] = None

    @classmethod
    def from_dict(cls, json_dict: Dict) -> LabelRowMetadata:
        created_at = json_dict.get("created_at", None)
        if created_at is not None:
            created_at = parse_datetime(created_at)
        last_edited_at = json_dict.get("last_edited_at", None)
        if last_edited_at is not None:
            last_edited_at = parse_datetime(last_edited_at)

        annotation_task_status = (
            AnnotationTaskStatus(json_dict["annotation_task_status"])
            if json_dict.get("annotation_task_status", None) is not None
            else None
        )
        return LabelRowMetadata(
            label_hash=json_dict.get("label_hash", None),
            created_at=created_at,
            last_edited_at=last_edited_at,
            data_hash=json_dict["data_hash"],
            group_hash=json_dict.get("group_hash", None),
            dataset_hash=json_dict["dataset_hash"],
            dataset_title=json_dict["dataset_title"],
            data_title=json_dict["data_title"],
            data_type=json_dict["data_type"],
            data_link=json_dict["data_link"],
            label_status=LabelStatus(json_dict["label_status"]),
            annotation_task_status=annotation_task_status,
            workflow_graph_node=WorkflowGraphNode.from_optional_dict(json_dict.get("workflow_graph_node")),
            is_shadow_data=json_dict.get("is_shadow_data", False),
            number_of_frames=json_dict["number_of_frames"],
            duration=json_dict.get("duration", None),
            frames_per_second=json_dict.get("frames_per_second", None),
            audio_codec=json_dict.get("audio_codec", None),
            audio_bit_depth=json_dict.get("audio_bit_depth", None),
            audio_num_channels=json_dict.get("audio_num_channels", None),
            audio_sample_rate=json_dict.get("audio_sample_rate", None),
            height=json_dict.get("height"),
            width=json_dict.get("width"),
            task_uuid=UUID(json_dict["task_uuid"]) if json_dict.get("task_uuid", None) is not None else None,
            priority=json_dict.get("priority"),
            client_metadata=json_dict.get("client_metadata", None),
            images_data=json_dict.get("images_data", None),
            file_type=json_dict.get("file_type"),
            is_valid=bool(json_dict.get("is_valid", True)),
            branch_name=json_dict["branch_name"],
            backing_item_uuid=UUID(json_dict["backing_item_uuid"])
            if json_dict.get("backing_item_uuid") is not None
            else None,
            assigned_user_email=json_dict.get("assigned_user_email", None),
            last_actioned_by_user_email=json_dict.get("last_actioned_by_user_email", None),
        )

    @classmethod
    def from_list(cls, json_list: list) -> List[LabelRowMetadata]:
        ret = []
        for i in json_list:
            ret.append(cls.from_dict(i))
        return ret

    def to_dict(self) -> Dict:
        """Returns:
        The dict equivalent of LabelRowMetadata.
        """

        def transform(value: Any):
            if isinstance(value, Enum):
                return value.value
            elif isinstance(value, datetime.datetime):
                return value.isoformat()
            return value

        return {k: transform(v) for k, v in asdict(self).items()}


class LabelValidationState(BaseDTO):
    label_hash: str
    branch_name: str
    version: int
    is_valid: bool
    errors: List[str]


class WorkflowGraphNodeDTO(BaseDTO):
    uuid: str
    title: str


class LabelRowMetadataDTO(BaseDTO):
    """Contains helpful information about a label row."""

    label_hash: Optional[str] = Field(default=None, alias="label_uuid")
    """Only present if the label row is initiated"""
    created_at: Optional[datetime.datetime] = None
    """Only present if the label row is initiated"""
    last_edited_at: Optional[datetime.datetime] = None
    """Only present if the label row is initiated"""
    branch_name: str
    """Only present if the label row is initiated or branch_name is set specifically"""

    data_hash: str = Field(alias="data_uuid")
    group_hash: Optional[str] = Field(default=None, alias="group_uuid")
    dataset_hash: str = Field(alias="dataset_uuid")
    dataset_title: str
    data_title: str
    data_type: str
    data_link: Optional[str] = None

    """Can be `None` for label rows of image groups or DICOM series."""
    label_status: LabelStatus
    """Can be `None` for TMS2 projects"""
    annotation_task_status: Optional[AnnotationTaskStatus] = None
    """Only available for TMS2 project"""
    workflow_graph_node: Optional[WorkflowGraphNode] = None
    is_shadow_data: bool = False

    """Only available for the VIDEO and AUDIO data_type"""
    frames_per_second: Optional[float]
    number_of_frames: int
    duration: Optional[float]

    """Only available for the VIDEO data_type"""
    height: Optional[int]
    width: Optional[int]

    """Only available for the AUDIO data_type"""
    audio_sample_rate: Optional[int]
    audio_codec: Optional[str]
    audio_bit_depth: Optional[int]
    audio_num_channels: Optional[int]

    task_uuid: Optional[UUID] = None
    priority: Optional[float] = None
    """Only available for not complete tasks"""
    client_metadata: Optional[Dict[str, Any]] = None
    images_data: Optional[List[Any]] = None
    file_type: Optional[str] = None
    """Only available for certain read requests"""
    is_valid: bool = True

    backing_item_uuid: Optional[UUID] = None
    assigned_user_email: Optional[str] = None
    last_actioned_by_user_email: Optional[str] = None


def label_row_metadata_dto_to_label_row_metadata(label_row_metadata_dto: LabelRowMetadataDTO) -> LabelRowMetadata:
    return LabelRowMetadata(
        label_hash=label_row_metadata_dto.label_hash,
        created_at=label_row_metadata_dto.created_at,
        last_edited_at=label_row_metadata_dto.last_edited_at,
        data_hash=label_row_metadata_dto.data_hash,
        group_hash=label_row_metadata_dto.group_hash,
        dataset_hash=label_row_metadata_dto.dataset_hash,
        dataset_title=label_row_metadata_dto.dataset_title,
        data_title=label_row_metadata_dto.data_title,
        data_type=label_row_metadata_dto.data_type,
        data_link=label_row_metadata_dto.data_link,
        label_status=label_row_metadata_dto.label_status,
        annotation_task_status=label_row_metadata_dto.annotation_task_status,
        workflow_graph_node=label_row_metadata_dto.workflow_graph_node,
        is_shadow_data=label_row_metadata_dto.is_shadow_data,
        number_of_frames=label_row_metadata_dto.number_of_frames,
        duration=label_row_metadata_dto.duration,
        audio_sample_rate=label_row_metadata_dto.audio_sample_rate,
        audio_codec=label_row_metadata_dto.audio_codec,
        audio_bit_depth=label_row_metadata_dto.audio_bit_depth,
        audio_num_channels=label_row_metadata_dto.audio_num_channels,
        frames_per_second=label_row_metadata_dto.frames_per_second,
        height=label_row_metadata_dto.height,
        width=label_row_metadata_dto.width,
        task_uuid=label_row_metadata_dto.task_uuid,
        priority=label_row_metadata_dto.priority,
        client_metadata=label_row_metadata_dto.client_metadata,
        images_data=label_row_metadata_dto.images_data,
        file_type=label_row_metadata_dto.file_type,
        is_valid=label_row_metadata_dto.is_valid,
        branch_name=label_row_metadata_dto.branch_name,
        backing_item_uuid=label_row_metadata_dto.backing_item_uuid,
        assigned_user_email=label_row_metadata_dto.assigned_user_email,
        last_actioned_by_user_email=label_row_metadata_dto.last_actioned_by_user_email,
    )



from enum import Enum
from typing import Optional

from encord.orm.base_dto import BaseDTO


class ActiveProjectMode(Enum):
    DATA = "data"
    LABEL = "label"
    METRIC = "metric"
    ADVANCED = "advanced"


class ActiveProjectImportPayload(BaseDTO):
    project_mode: ActiveProjectMode
    video_sampling_rate: Optional[float] = None



"""---
title: "Objects - Attributes"
slug: "sdk-ref-objects-attributes"
hidden: false
metadata:
  title: "Objects - Attributes"
  description: "Encord SDK Objects - Attributes class."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Dict, Generic, List, Optional, Sequence, Type, TypeVar

from encord.objects.ontology_element import (
    NestedID,
    OntologyElement,
    OntologyNestedElement,
    _build_identifiers,
    _decode_nested_uid,
    _get_element_by_hash,
    _nested_id_from_json_str,
)

OptionType = TypeVar("OptionType", bound="Option")
AttributeType = TypeVar("AttributeType", bound="Attribute")


class Attribute(OntologyNestedElement, Generic[OptionType]):
    """Base class for shared Attribute fields"""

    name: str
    required: bool
    archived: bool
    dynamic: bool
    """
    The `dynamic` member is part of every attribute. However it can only be true for top level (not nested) attributes
    that are part of an :class:`encord.objects.ontology_object.Object`.
    """

    def __init__(self, uid: NestedID, feature_node_hash: str, name: str, required: bool, archived: bool, dynamic: bool):
        super().__init__(uid=uid, feature_node_hash=feature_node_hash)
        self.name = name
        self.required = required
        self.archived = archived
        self.dynamic = dynamic

    @property
    def options(self) -> Sequence[OptionType]:
        return []

    @property
    def title(self) -> str:
        return self.name

    @staticmethod
    @abstractmethod
    def get_property_type() -> PropertyType:
        pass

    @abstractmethod
    def _encode_options(self) -> Optional[List[dict]]:
        pass

    def to_dict(self) -> Dict[str, Any]:
        ret = self._encode_base()

        options = self._encode_options()
        if options is not None:
            ret["options"] = options

        return ret

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Attribute:
        property_type = d["type"]
        common_attribute_fields = cls._decode_common_attribute_fields(d)
        if property_type == RadioAttribute.get_property_type():
            return RadioAttribute(
                **common_attribute_fields,
                options=[NestableOption.from_dict(x) for x in d.get("options", [])],
            )

        elif property_type == ChecklistAttribute.get_property_type():
            return ChecklistAttribute(
                **common_attribute_fields,
                options=[FlatOption.from_dict(x) for x in d.get("options", [])],
            )

        elif property_type == TextAttribute.get_property_type():
            return TextAttribute(
                **common_attribute_fields,
            )
        elif property_type == NumericAttribute.get_property_type():
            return NumericAttribute(
                **common_attribute_fields,
            )

        raise TypeError(
            f"Attribute is ill-formed: '{d}'. Expected to see either "
            f"attribute specific fields or option specific fields. Got both or none of them."
        )

    def _encode_base(self) -> Dict[str, Any]:
        ret: Dict[str, Any] = dict()
        ret["id"] = _decode_nested_uid(self.uid)
        ret["name"] = self.name
        ret["type"] = self.get_property_type().value
        ret["featureNodeHash"] = self.feature_node_hash
        ret["required"] = self.required
        ret["archived"] = self.archived
        ret["dynamic"] = self.dynamic

        return ret

    @staticmethod
    def _decode_common_attribute_fields(attribute_dict: Dict[str, Any]) -> Dict[str, Any]:
        return {
            "uid": _nested_id_from_json_str(attribute_dict["id"]),
            "feature_node_hash": attribute_dict["featureNodeHash"],
            "name": attribute_dict["name"],
            "required": attribute_dict["required"],
            "dynamic": attribute_dict.get("dynamic", False),
            "archived": attribute_dict.get("archived", False),
        }

    def __eq__(self, other: object):
        return (
            isinstance(other, Attribute) and self.uid == other.uid and self.feature_node_hash == other.feature_node_hash
        )


class RadioAttribute(Attribute["NestableOption"]):
    _options: List[NestableOption]

    def __init__(
        self,
        uid: NestedID,
        feature_node_hash: str,
        name: str,
        required: bool,
        archived: bool,
        dynamic: bool,
        options: Optional[List[NestableOption]] = None,
    ):
        super().__init__(
            uid=uid,
            feature_node_hash=feature_node_hash,
            name=name,
            required=required,
            archived=archived,
            dynamic=dynamic,
        )
        self._options = options if options is not None else []

    @property
    def options(self) -> Sequence[NestableOption]:
        return self._options

    @property
    def children(self) -> Sequence[OntologyElement]:
        return self._options

    @staticmethod
    def get_property_type() -> PropertyType:
        return PropertyType.RADIO

    def _encode_options(self) -> Optional[List[Dict[str, Any]]]:
        if len(self._options) == 0:
            return None
        return [option.to_dict() for option in self._options]

    def add_option(
        self,
        label: str,
        value: Optional[str] = None,
        local_uid: Optional[int] = None,
        feature_node_hash: Optional[str] = None,
    ) -> NestableOption:
        """Args:
            label: user-visible name of the option
            value: internal unique value; optional; normally mechanically constructed from the label
            local_uid: integer identifier of the option. Normally auto-generated;
                    omit this unless the aim is to create an exact clone of existing ontology
            feature_node_hash: global identifier of the option. Normally auto-generated;
                    omit this unless the aim is to create an exact clone of existing ontology

        Returns:
            a `NestableOption` instance attached to the attribute. This can be further specified by adding nested attributes.
        """
        return _add_option(self._options, NestableOption, label, self.uid, local_uid, feature_node_hash, value)


class ChecklistAttribute(Attribute["FlatOption"]):
    _options: List[FlatOption]

    def __init__(
        self,
        uid: NestedID,
        feature_node_hash: str,
        name: str,
        required: bool,
        archived: bool,
        dynamic: bool,
        options: Optional[List[FlatOption]] = None,
    ):
        super().__init__(
            uid=uid,
            feature_node_hash=feature_node_hash,
            name=name,
            required=required,
            archived=archived,
            dynamic=dynamic,
        )
        self._options = options if options is not None else []

    @staticmethod
    def get_property_type() -> PropertyType:
        return PropertyType.CHECKLIST

    def _encode_options(self) -> Optional[List[Dict[str, Any]]]:
        if len(self._options) == 0:
            return None
        return [option.to_dict() for option in self._options]

    @property
    def options(self) -> Sequence[FlatOption]:
        return self._options

    @property
    def children(self) -> Sequence[OntologyElement]:
        return self._options

    def add_option(
        self,
        label: str,
        value: Optional[str] = None,
        local_uid: Optional[int] = None,
        feature_node_hash: Optional[str] = None,
    ) -> FlatOption:
        """Args:
            label: user-visible name of the option
            value: internal unique value; optional; normally mechanically constructed from the label
            local_uid: integer identifier of the option. Normally auto-generated;
                    omit this unless the aim is to create an exact clone of existing ontology
            feature_node_hash: global identifier of the option. Normally auto-generated;
                    omit this unless the aim is to create an exact clone of existing ontology
        Returns:
            a `FlatOption` instance attached to the attribute.
        """
        return _add_option(self._options, FlatOption, label, self.uid, local_uid, feature_node_hash, value)


class TextAttribute(Attribute["FlatOption"]):
    def __init__(self, uid: NestedID, feature_node_hash: str, name: str, required: bool, archived: bool, dynamic: bool):
        super().__init__(
            uid=uid,
            feature_node_hash=feature_node_hash,
            name=name,
            required=required,
            archived=archived,
            dynamic=dynamic,
        )

    @staticmethod
    def get_property_type() -> PropertyType:
        return PropertyType.TEXT

    def _encode_options(self) -> Optional[List[Dict[str, Any]]]:
        return None


class NumericAttribute(Attribute["FlatOption"]):
    def __init__(self, uid: NestedID, feature_node_hash: str, name: str, required: bool, archived: bool, dynamic: bool):
        super().__init__(
            uid=uid,
            feature_node_hash=feature_node_hash,
            name=name,
            required=required,
            archived=archived,
            dynamic=dynamic,
        )

    @staticmethod
    def get_property_type() -> PropertyType:
        return PropertyType.NUMERIC

    def _encode_options(self) -> Optional[List[Dict[str, Any]]]:
        # Numeric attribute does not have a notion of "options"
        return None


def attribute_from_dict(d: Dict[str, Any]) -> Attribute:
    """Convenience functions as you cannot call static member on union types."""
    return Attribute.from_dict(d)


def _add_attribute(
    attributes: List[Attribute],
    cls: Type[AttributeType],
    name: str,
    parent_uid: List[int],
    local_uid: Optional[int] = None,
    feature_node_hash: Optional[str] = None,
    required: bool = False,
    dynamic: bool = False,
) -> AttributeType:
    local_uid, feature_node_hash = _build_identifiers(attributes, local_uid, feature_node_hash)
    attr = cls(
        name=name,
        uid=parent_uid + [local_uid],
        feature_node_hash=feature_node_hash,
        required=required,
        archived=False,
        dynamic=dynamic,
    )

    attributes.append(attr)
    return attr


def _get_attribute_by_hash(feature_node_hash: str, attributes: List[Attribute]) -> Optional[Attribute]:
    return _get_element_by_hash(feature_node_hash, attributes, type_=Attribute)


def attributes_to_list_dict(attributes: List[Attribute]) -> List[Dict[str, Any]]:
    return [attribute.to_dict() for attribute in attributes]


from encord.objects.common import PropertyType
from encord.objects.options import (  # pylint: disable=unused-import
    FlatOption,
    NestableOption,
    Option,
    _add_option,
)



from encord.workflow.stages.agent import AgentStage
from encord.workflow.stages.annotation import AnnotationStage, AnnotationTask, AnnotationTaskStatus
from encord.workflow.stages.consensus_annotation import ConsensusAnnotationStage, ConsensusAnnotationTask
from encord.workflow.stages.consensus_review import ConsensusReviewStage, ConsensusReviewTask, ConsensusReviewTaskStatus
from encord.workflow.stages.final import FinalStage, FinalStageTask
from encord.workflow.stages.review import ReviewStage, ReviewTask, ReviewTaskStatus
from encord.workflow.workflow import Workflow

__all__ = [
    "AgentStage",
    "AnnotationStage",
    "AnnotationTask",
    "AnnotationTaskStatus",
    "ConsensusAnnotationStage",
    "ConsensusAnnotationTask",
    "ConsensusReviewStage",
    "ConsensusReviewTask",
    "ConsensusReviewTaskStatus",
    "FinalStage",
    "FinalStageTask",
    "ReviewStage",
    "ReviewTask",
    "ReviewTaskStatus",
    "Workflow",
]



"""---
title: "Enums"
slug: "sdk-ref-enums"
hidden: false
metadata:
  title: "Enums"
  description: "Encord SDK Enums."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from typing import Any

from encord.common.enum import StringEnum


class DataType(StringEnum):
    VIDEO = "video"
    IMG_GROUP = "img_group"
    DICOM = "dicom"
    IMAGE = "image"
    DICOM_STUDY = "dicom_study"
    NIFTI = "nifti"
    AUDIO = "audio"
    PLAIN_TEXT = "plain_text"
    PDF = "pdf"
    GROUP = "group"
    SCENE = "scene"

    # will be displayed if the Encord platform has a new data type that is not present in this SDK version. Please upgrade your SDK version
    MISSING_DATA_TYPE = "_MISSING_DATA_TYPE_"

    @classmethod
    def _missing_(cls, value: Any) -> DataType:
        return cls.MISSING_DATA_TYPE

    @staticmethod
    def from_upper_case_string(string: str) -> DataType:
        for data_type in DataType:
            if string == data_type.to_upper_case_string():
                return data_type

        return DataType.MISSING_DATA_TYPE

    def to_upper_case_string(self) -> str:
        return self.value.upper()


GEOMETRIC_TYPES = {
    DataType.VIDEO,
    DataType.IMAGE,
    DataType.IMG_GROUP,
    DataType.GROUP,
    DataType.DICOM,
    DataType.DICOM_STUDY,
    DataType.NIFTI,
    DataType.PDF,
    DataType.SCENE,
}


def is_geometric(data_type: DataType) -> bool:
    return data_type in GEOMETRIC_TYPES



from __future__ import annotations

from typing import List, TypedDict, Union


class Answer(TypedDict):
    name: str
    value: str
    featureHash: str


class ClassificationObject(TypedDict):
    name: str
    value: str
    answers: Union[List[Answer], str, float]
    featureHash: str
    manualAnnotation: bool


class ClassificationAnswer(TypedDict):
    classificationHash: str
    featureHash: str
    classifications: List[ClassificationObject]
    range: Union[List[List[int]], None]

    # The values below are only guaranteed for non-geometric files
    confidence: Union[float, None]
    createdAt: Union[str, None]
    createdBy: Union[str, None]
    lastEditedAt: Union[str, None]
    lastEditedBy: Union[str, None]
    manualAnnotation: Union[bool, None]

    reviews: Union[List, None]
    """The reviews field is deprecated and soon to be removed"""


def is_containing_metadata(answer: ClassificationAnswer) -> bool:
    """Check if the classification answer contains necessary metadata fields."""
    return answer.get("createdBy") is not None



from enum import Enum, auto
from typing import List, Optional, Union
from uuid import UUID

from encord.orm.analytics import CamelStrEnum
from encord.orm.base_dto import BaseDTO


class DicomDeIdRedactTextMode(CamelStrEnum):
    REDACT_ALL_TEXT = auto()
    REDACT_NO_TEXT = auto()
    REDACT_SENSITIVE_TEXT = auto()


class DicomDeIdSaveConditionType(CamelStrEnum):
    NOT_SUBSTR = auto()
    IN = auto()


class DicomDeIdSaveCondition(BaseDTO):
    value: Union[str, List[str]]
    condition_type: DicomDeIdSaveConditionType
    dicom_tag: str


class DicomDeIdStartPayload(BaseDTO):
    integration_uuid: UUID
    dicom_urls: List[str]
    redact_dicom_tags: bool = True
    redact_pixels_mode: DicomDeIdRedactTextMode = DicomDeIdRedactTextMode.REDACT_NO_TEXT
    save_conditions: Optional[List[DicomDeIdSaveCondition]] = None
    upload_dir: Optional[str] = None


class DicomDeIdGetResultParams(BaseDTO):
    timeout_seconds: int


class DicomDeIdGetResultLongPollingStatus(str, Enum):
    DONE = "DONE"
    ERROR = "ERROR"
    PENDING = "PENDING"


class DicomDeIdGetResultResponse(BaseDTO):
    status: DicomDeIdGetResultLongPollingStatus
    urls: Optional[List[str]] = None



"""---
title: "Review Stage"
slug: "sdk-ref-stage-review"
hidden: false
metadata:
  title: "Review Stage"
  description: "Encord SDK Review Stage."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from enum import Enum
from typing import Iterable, List, Literal, Optional, Tuple, Union
from uuid import UUID

from encord.common.utils import ensure_list, ensure_uuid_list
from encord.http.bundle import Bundle
from encord.issues.issue_client import TaskIssues
from encord.orm.base_dto import BaseDTO, Field, PrivateAttr
from encord.orm.workflow import WorkflowStageType
from encord.workflow.common import (
    TasksQueryParams,
    WorkflowAction,
    WorkflowClient,
    WorkflowReviewAction,
    WorkflowStageBase,
    WorkflowTask,
)


class LabelReviewStatus(str, Enum):
    NEW = "NEW"
    APPROVED = "APPROVED"
    REJECTED = "REJECTED"
    RESOLVED = "RESOLVED"
    REOPENED = "REOPENED"


class _LabelReviewActionApprove(WorkflowReviewAction):
    action: Literal["APPROVE"] = "APPROVE"


class _LabelReviewActionReject(WorkflowReviewAction):
    action: Literal["REJECT"] = "REJECT"
    comment: Optional[str] = None
    comment_tags: Optional[List[str]] = None


class _LabelReviewActionReopen(WorkflowReviewAction):
    action: Literal["REOPEN"] = "REOPEN"


class LabelReview(BaseDTO):
    _workflow_client: Optional[WorkflowClient] = PrivateAttr(None)
    _stage_uuid: Optional[UUID] = PrivateAttr(None)
    _task_uuid: Optional[UUID] = PrivateAttr(None)

    uuid: UUID = Field(alias="reviewUuid")
    status: LabelReviewStatus

    label_type: str
    label_id: str

    def _get_client_data(self) -> Tuple[WorkflowClient, UUID, UUID]:
        assert self._workflow_client
        assert self._stage_uuid
        assert self._task_uuid
        return self._workflow_client, self._stage_uuid, self._task_uuid

    def approve(self, *, bundle: Optional[Bundle] = None):
        """Approves the review.

        **Parameters**

        - `bundle` (Optional[Bundle]): Optional bundle parameter.
        """
        workflow_client, stage_uuid, task_uuid = self._get_client_data()
        workflow_client.label_review_action(
            stage_uuid, task_uuid, _LabelReviewActionApprove(review_uuid=self.uuid), bundle=bundle
        )

    def reject(
        self,
        *,
        comment: Optional[str] = None,
        issue_tags: Optional[List[str]] = None,
        bundle: Optional[Bundle] = None,
    ):
        """Rejects the review.

        **Parameters**

        - `bundle` (Optional[Bundle]): Optional bundle parameter.
        """
        workflow_client, stage_uuid, task_uuid = self._get_client_data()
        workflow_client.label_review_action(
            stage_uuid,
            task_uuid,
            _LabelReviewActionReject(review_uuid=self.uuid, comment=comment, comment_tags=issue_tags),
            bundle=bundle,
        )

    def reopen(self, *, bundle: Optional[Bundle] = None):
        """Reopens the review.

        **Parameters**

        - `bundle` (Optional[Bundle]): Optional bundle parameter.
        """
        workflow_client, stage_uuid, task_uuid = self._get_client_data()
        workflow_client.label_review_action(
            stage_uuid, task_uuid, _LabelReviewActionReopen(review_uuid=self.uuid), bundle=bundle
        )


class ReviewTaskStatus(str, Enum):
    NEW = "NEW"
    ASSIGNED = "ASSIGNED"
    RELEASED = "RELEASED"
    REOPENED = "REOPENED"


class _ReviewTasksQueryParams(TasksQueryParams):
    user_emails: Optional[List[str]] = None
    data_hashes: Optional[List[UUID]] = None
    dataset_hashes: Optional[List[UUID]] = None
    data_title_contains: Optional[str] = None
    statuses: Optional[List[ReviewTaskStatus]] = None


class ReviewStage(WorkflowStageBase):
    stage_type: Literal[WorkflowStageType.REVIEW] = WorkflowStageType.REVIEW

    """
    The Review stage for Workflows.

    This stage can appear in Consensus and non-Consensus Workflows.
    """

    def get_tasks(
        self,
        *,
        assignee: Union[List[str], str, None] = None,
        data_hash: Union[List[UUID], UUID, List[str], str, None] = None,
        dataset_hash: Union[List[UUID], UUID, List[str], str, None] = None,
        data_title: Optional[str] = None,
        status: Union[ReviewTaskStatus, List[ReviewTaskStatus], None] = None,
    ) -> Iterable[ReviewTask]:
        """Retrieves tasks for the ReviewStage.

        **Parameters**

        - `assignee` (Union[List[str], str, None]): User assigned to a task.
        - `data_hash` (Union[List[UUID], UUID, List[str], str, None]): Unique ID for the data unit.
        - `dataset_hash` (Union[List[UUID], UUID, List[str], str, None]): Unique ID for the dataset that the data unit belongs to.
        - `data_title` (Optional[str]): Name of the data unit.
        - `status` (Union[ReviewTaskStatus, List[ReviewTaskStatus], None]): Status of the task.

        **Returns**

        An iterable of `ReviewTask` instances with the following information:
        - `uuid`: Unique identifier for the task.
        - `created_at`: Time and date the task was created.
        - `updated_at`: Time and date the task was last edited.
        - `assignee`: The user currently assigned to the task. The value is None if no one is assigned to the task.
        - `data_hash`: Unique identifier for the data unit.
        - `data_title`: Name/title of the data unit.
        """
        params = _ReviewTasksQueryParams(
            user_emails=ensure_list(assignee),
            data_hashes=ensure_uuid_list(data_hash),
            dataset_hashes=ensure_uuid_list(dataset_hash),
            data_title_contains=data_title,
            statuses=ensure_list(status),
        )

        for task in self._workflow_client.get_tasks(self.uuid, params, type_=ReviewTask):
            task._stage_uuid = self.uuid
            task._workflow_client = self._workflow_client
            task._task_issues = TaskIssues(
                api_client=self._workflow_client.api_client,
                project_uuid=self._workflow_client.project_hash,
                data_uuid=task.data_hash,
            )
            yield task


class _ActionApprove(WorkflowAction):
    action: Literal["APPROVE"] = "APPROVE"
    approve_label_reviews: bool = True
    assignee: Optional[str] = None
    retain_assignee: bool = False


class _ActionReject(WorkflowAction):
    action: Literal["REJECT"] = "REJECT"
    reject_label_reviews: bool = True
    assignee: Optional[str] = None
    retain_assignee: bool = False


class _ActionAssign(WorkflowAction):
    action: Literal["ASSIGN"] = "ASSIGN"
    assignee: str


class _ActionRelease(WorkflowAction):
    action: Literal["RELEASE"] = "RELEASE"


class ReviewTask(WorkflowTask):
    """Represents tasks in non-Consensus Review stages.

    **Attributes**

    - `status` (ReviewTaskStatus): Status of the review task.
    - `data_hash` (UUID): Unique identifier for the data unit.
    - `data_title` (str): Name of the data unit.
    - `assignee` (Optional[str]): User assigned to the task.

    **Allowed actions**

    - `approve`: Approves a task.
    - `reject`: Rejects a task.
    - `assign`: Assigns a task to a user.
    - `release`: Releases a task from the current user.
    - `move`: Moves the review task to another stage in the workflow.
    """

    status: ReviewTaskStatus
    data_hash: UUID
    data_title: str
    assignee: Optional[str]

    def approve(
        self,
        *,
        assignee: Optional[str] = None,
        retain_assignee: bool = False,
        bundle: Optional[Bundle] = None,
    ) -> None:
        """Approves the task.

        **Parameters**

        - `assignee` (Optional[str]): User email to be assigned to the task whilst approving the task.
        - `retain_assignee` (bool): Retains the current assignee whilst approving the task. This is ignored if `assignee` is provided. An error will occur if the task does not already have an assignee and `retain_assignee` is True.
        - `bundle` (Optional[Bundle]): Optional bundle parameter.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(
            stage_uuid,
            _ActionApprove(task_uuid=self.uuid, assignee=assignee, retain_assignee=retain_assignee),
            bundle=bundle,
        )

    def reject(
        self,
        *,
        assignee: Optional[str] = None,
        retain_assignee: bool = False,
        bundle: Optional[Bundle] = None,
    ) -> None:
        """Rejects the task.

        **Parameters**

        - `assignee` (Optional[str]): User email to be assigned to the task whilst rejecting the task.
        - `retain_assignee` (bool): Retains the current assignee whilst rejecting the task. This is ignored if `assignee` is provided. An error will occur if the task does not already have an assignee and `retain_assignee` is True.
        - `bundle` (Optional[Bundle]): Optional bundle parameter.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(
            stage_uuid,
            _ActionReject(task_uuid=self.uuid, assignee=assignee, retain_assignee=retain_assignee),
            bundle=bundle,
        )

    def assign(self, assignee: str, *, bundle: Optional[Bundle] = None) -> None:
        """Assigns the task to a user.

        **Parameters**

        - `assignee` (str): The user to assign the task to.
        - `bundle` (Optional[Bundle]): Optional bundle parameter.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(
            stage_uuid,
            _ActionAssign(task_uuid=self.uuid, assignee=assignee),
            bundle=bundle,
        )

    def release(self, *, bundle: Optional[Bundle] = None) -> None:
        """Releases the task from the current user.

        **Parameters**

        - `bundle` (Optional[Bundle]): Optional bundle parameter.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(
            stage_uuid,
            _ActionRelease(task_uuid=self.uuid),
            bundle=bundle,
        )

    def move(self, *, destination_stage_uuid: UUID, bundle: Optional[Bundle] = None) -> None:
        """Moves the review task from its current stage to another stage.

        **Parameters**

        - `destination_stage_uuid` (UUID): Unique identifier of the stage to move the task to.
        - `bundle` (Optional[Bundle]): Optional bundle parameter.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.move(
            origin_stage_uuid=stage_uuid,
            destination_stage_uuid=destination_stage_uuid,
            task_uuids=[self.uuid],
            bundle=bundle,
        )

    def get_label_reviews(
        self, status: Union[ReviewTaskStatus, List[ReviewTaskStatus], None] = None
    ) -> Iterable[LabelReview]:
        """Retrieves label reviews for the Review task.

        **Parameters**
        - `status` (Union[ReviewTaskStatus, List[ReviewTaskStatus], None]): Status of the task.

        **Returns**

        An iterable of `ReviewTask` instances with the following information:
        - `uuid`: Unique identifier for label review.
        - `status`: Current status of the label review.
        - `label_type`: Type of the label. Can be either Object or Classification.
        - `label_id`: Unique identifier of the label.
        """
        workflow_client, stage_uuid = self._get_client_data()
        for r in workflow_client.get_label_reviews(stage_uuid, self.uuid, type_=LabelReview):
            r._workflow_client = self._workflow_client
            r._stage_uuid = self._stage_uuid
            r._task_uuid = self.uuid
            yield r

    @property
    def issues(self) -> TaskIssues:
        """Returns the issue client for the task."""
        assert self._task_issues
        return self._task_issues



from typing import List, Optional
from uuid import UUID

from encord.orm.base_dto import BaseDTO


class CloudIntegration(BaseDTO):
    id: str
    title: str


class CloudIntegrationV2(BaseDTO):
    integration_uuid: UUID
    title: str


class GetCloudIntegrationsResponse(BaseDTO):
    result: List[CloudIntegrationV2]


class GetCloudIntegrationsParams(BaseDTO):
    filter_integration_uuids: Optional[List[UUID]] = None
    filter_integration_titles: Optional[List[str]] = None
    include_org_access: bool = False



"""---
title: "Workflow"
slug: "sdk-ref-workflows"
hidden: false
metadata:
  title: "Workflows"
  description: "Encord SDK Workflow class."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from typing import List, Optional, Type, TypeVar, Union
from uuid import UUID

from typing_extensions import Annotated

from encord.http.v2.api_client import ApiClient
from encord.objects.utils import checked_cast
from encord.orm.base_dto import Field
from encord.orm.workflow import WorkflowAgentNode, WorkflowDTO, WorkflowNode, WorkflowStageType
from encord.workflow.common import WorkflowClient
from encord.workflow.stages.agent import AgentPathway, AgentStage
from encord.workflow.stages.annotation import AnnotationStage
from encord.workflow.stages.consensus_annotation import ConsensusAnnotationStage
from encord.workflow.stages.consensus_review import ConsensusReviewStage
from encord.workflow.stages.final import FinalStage
from encord.workflow.stages.review import ReviewStage

WorkflowStage = Annotated[
    Union[AnnotationStage, ReviewStage, ConsensusAnnotationStage, ConsensusReviewStage, FinalStage, AgentStage],
    Field(discriminator="stage_type"),
]


def _construct_stage(workflow_client: WorkflowClient, node: Union[WorkflowAgentNode, WorkflowNode]) -> WorkflowStage:
    if node.stage_type == WorkflowStageType.ANNOTATION:
        return AnnotationStage(uuid=node.uuid, title=node.title, _workflow_client=workflow_client)
    elif node.stage_type == WorkflowStageType.REVIEW:
        return ReviewStage(uuid=node.uuid, title=node.title, _workflow_client=workflow_client)
    elif node.stage_type == WorkflowStageType.CONSENSUS_ANNOTATION:
        return ConsensusAnnotationStage(uuid=node.uuid, title=node.title, _workflow_client=workflow_client)
    elif node.stage_type == WorkflowStageType.CONSENSUS_REVIEW:
        return ConsensusReviewStage(uuid=node.uuid, title=node.title, _workflow_client=workflow_client)
    elif node.stage_type == WorkflowStageType.DONE:
        return FinalStage(uuid=node.uuid, title=node.title, _workflow_client=workflow_client)
    elif node.stage_type == WorkflowStageType.AGENT:
        assert isinstance(node, WorkflowAgentNode)
        return AgentStage(
            uuid=node.uuid,
            title=node.title,
            pathways=[
                AgentPathway(uuid=pathway.uuid, name=pathway.title, destination_uuid=pathway.destination_uuid)
                for pathway in node.pathways
            ],
            _workflow_client=workflow_client,
        )
    else:
        raise AssertionError(f"Unknown stage type: {node.stage_type}")


WorkflowStageT = TypeVar(
    "WorkflowStageT",
    AnnotationStage,
    ReviewStage,
    ConsensusAnnotationStage,
    ConsensusReviewStage,
    FinalStage,
    AgentStage,
)


def _ensure_uuid(v: Union[str, UUID]) -> UUID:
    if isinstance(v, UUID):
        return v
    else:
        return UUID(v)


class Workflow:
    stages: List[WorkflowStage] = []

    """
    Workflow class in Projects.
    """

    def __init__(self, api_client: ApiClient, project_hash: UUID, workflow_orm: WorkflowDTO):
        workflow_client = WorkflowClient(api_client, project_hash)

        self.stages = [_construct_stage(workflow_client, stage) for stage in workflow_orm.stages]

    def get_stage(
        self,
        *,
        name: Optional[str] = None,
        uuid: Optional[Union[UUID, str]] = None,
        type_: Optional[Type[WorkflowStageT]] = None,
    ) -> WorkflowStageT:
        """**Params**

        - name: Name of the stage.
        - uuid: Unique identifier for the stage.
        - type_: The type of stage.

        **Returns**

        Returns a Workflow stage (`type_`) from non-Consensus and Consensus Projects.
        """
        for stage in self.stages:
            if (uuid is not None and stage.uuid == _ensure_uuid(uuid)) or (name is not None and stage.title == name):
                if type_ is not None and not isinstance(stage, type_):
                    raise AssertionError(
                        f"Expected '{type_.__name__}' but got '{type(stage).__name__}' for the requested workflow stage '{uuid or name}'"
                    )
                return checked_cast(stage, type_)

        raise ValueError(f"No matching stage found: '{uuid or name}'")



from __future__ import annotations

from datetime import datetime
from enum import IntEnum
from typing import Dict, Optional

# pylint: disable=unused-import
from encord.objects.ontology_structure import OntologyStructure
from encord.orm.base_dto import BaseDTO, dto_validator
from encord.orm.formatter import Formatter
from encord.utilities.ontology_user import OntologyUserRole


class Ontology(dict, Formatter):
    def __init__(
        self,
        title: str,
        structure: OntologyStructure,
        ontology_hash: str,
        created_at: datetime,
        last_edited_at: datetime,
        description: Optional[str] = None,
        user_role: Optional[OntologyUserRole] = None,
    ):
        """DEPRECATED - prefer using the :class:`encord.ontology.Ontology` class instead.

        This class has dict-style accessors for backwards compatibility.
        Clients who are using this class for the first time are encouraged to use the property accessors and setters
        instead of the underlying dictionary.
        The mixed use of the `dict` style member functions and the property accessors and setters is discouraged.

        WARNING: Do NOT use the `.data` member of this class. Its usage could corrupt the correctness of the
        datastructure.
        """
        super().__init__(
            {
                "ontology_hash": ontology_hash,
                "title": title,
                "description": description,
                "structure": structure,
                "created_at": created_at,
                "last_edited_at": last_edited_at,
                "user_role": user_role,
            }
        )

    @property
    def ontology_hash(self) -> str:
        return self["ontology_hash"]

    @property
    def title(self) -> str:
        return self["title"]

    @title.setter
    def title(self, value: str) -> None:
        self["title"] = value

    @property
    def description(self) -> str:
        return self["description"]

    @description.setter
    def description(self, value: str) -> None:
        self["description"] = value

    @property
    def structure(self) -> OntologyStructure:
        return self["structure"]

    @structure.setter
    def structure(self, value: OntologyStructure) -> None:
        self["structure"] = value

    @property
    def created_at(self) -> datetime:
        return self["created_at"]

    @property
    def last_edited_at(self) -> datetime:
        return self["last_edited_at"]

    @property
    def user_role(self) -> OntologyUserRole:
        return self["user_role"]

    @classmethod
    def from_dict(cls, json_dict: Dict) -> Ontology:
        return Ontology(
            title=json_dict["title"],
            description=json_dict["description"],
            ontology_hash=json_dict["ontology_hash"],
            structure=OntologyStructure.from_dict(json_dict["editor"]),
            created_at=json_dict["created_at"],
            last_edited_at=json_dict["last_edited_at"],
            user_role=json_dict.get("user_role"),  #  has to be like this to support the legacy endpoints for tests
        )


class CreateOrUpdateOntologyPayload(BaseDTO):
    title: str
    description: str
    editor: dict



from typing import Iterable, List, Optional, Set, Union, cast

from encord.objects.frames import Frames, Range, Ranges


class RangeManager:
    """Range Manager class to hold a list of frame ranges, and operate on them."""

    def __init__(self, frame_class: Optional[Frames] = None):
        self.ranges: Ranges = []
        if isinstance(frame_class, int):
            self.add_range(Range(start=frame_class, end=frame_class))
        elif isinstance(frame_class, Range):
            self.add_range(frame_class)
        elif isinstance(frame_class, (list, set)):
            if all(isinstance(x, int) for x in frame_class):
                for frame in frame_class:
                    self.add_range(Range(start=cast(int, frame), end=cast(int, frame)))
            elif all(isinstance(x, Range) for x in frame_class):
                self.add_ranges(cast(Ranges, frame_class))
        elif frame_class is None:
            self.ranges = []
        else:
            raise RuntimeError(f"Unexpected type for frames {type(frame_class)}.")

    def add_range(self, new_range: Range) -> None:
        """Add a range, merging any overlapping ranges."""
        if not self.ranges:
            self.ranges.append(new_range)
            return

        merged_ranges = []

        # Sort ranges based on the start of each range
        for existing_range in sorted(self.ranges, key=lambda r: r.start):
            if existing_range.overlaps(new_range):
                new_range.merge(existing_range)
            else:
                merged_ranges.append(existing_range)

        merged_ranges.append(new_range)  # Add the new (merged) range
        self.ranges = sorted(merged_ranges, key=lambda r: r.start)

    def add_ranges(self, new_ranges: Ranges) -> None:
        """Add multiple ranges."""
        for new_range in new_ranges:
            self.add_range(new_range)

    def remove_range(self, range_to_remove: Range) -> None:
        """Remove a specific range."""
        new_ranges = []

        for r in self.ranges:
            if not r.overlaps(range_to_remove):
                # No overlap
                new_ranges.append(r)
            else:
                # Partial overlap: split if needed
                if r.start < range_to_remove.start:
                    new_ranges.append(Range(r.start, range_to_remove.start - 1))
                if r.end > range_to_remove.end:
                    new_ranges.append(Range(range_to_remove.end + 1, r.end))

        self.ranges = new_ranges

    def remove_ranges(self, ranges_to_remove: Ranges) -> None:
        """Remove multiple ranges."""
        for r in ranges_to_remove:
            self.remove_range(r)

    def clear_ranges(self) -> None:
        """Clear all ranges."""
        self.ranges = []

    def get_ranges(self) -> Ranges:
        """Return the sorted list of merged ranges."""
        copied_ranges = [range.copy() for range in self.ranges]

        return sorted(copied_ranges, key=lambda r: r.start)

    def get_ranges_as_frames(self) -> Set[int]:
        """Returns set of intersecting frames"""
        res = set()
        for r in self.ranges:
            res.update(list(range(r.start, r.end + 1)))

        return res

    def intersection(self, other_frame_class: Frames) -> Ranges:
        """Returns list of intersecting ranges"""
        intersection_ranges: Ranges = []
        other_range_manager = RangeManager(other_frame_class)
        other_ranges = other_range_manager.get_ranges()
        current_ranges = self.get_ranges()

        # If either list of ranges is empty, there is no intersection
        if len(other_ranges) == 0 or len(current_ranges) == 0:
            return []

        # Since ranges are sorted, we can use 2-pointer method to find intersections
        current_index, other_index = 0, 0

        while current_index < len(current_ranges) and other_index < len(other_ranges):
            current_range = current_ranges[current_index]
            other_range = other_ranges[other_index]

            if current_range.overlaps(other_range):
                intersect_start = max(current_range.start, other_range.start)
                intersect_end = min(current_range.end, other_range.end)
                intersection_ranges.append(Range(intersect_start, intersect_end))

            # Move pointer for the range that ends first
            if current_range.end < other_range.end:
                current_index += 1
            else:
                other_index += 1

        return intersection_ranges
